"""Tauon Music Box - Misc Functions Module

TODO(Martin): Rewrite all tuple[int, int, int, int] things into actual ColorWhatever objects and pass the objects around instead of list[int]s
"""

# Copyright © 2015-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import colorsys
import glob
import locale
import logging
import math
import os
import random
import re
import shlex
import subprocess
import threading
import time
import urllib.parse
import zipfile
from collections import deque
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

from gi.repository import GLib

if TYPE_CHECKING:
	from collections.abc import Callable

	from tauon.t_modules.t_enums import QueueType
	from tauon.t_modules.t_main import TrackClass
	from tauon.t_modules.t_tagscan import TrackFile


@dataclass
class ColourRGBA:
	"""Red, Green, Blue and Alpha

	SDR, ranging from 0 to 255
	"""

	r: int
	g: int
	b: int
	a: int


@dataclass
class RadioStation:
	title: str
	stream_url: str
	country: str = ""
	website_url: str = ""
	icon: str = ""
	stream_url_fallback: str = ""


@dataclass
class RadioPlaylist:
	name: str
	uid: int
	scroll: int = 0
	stations: list[RadioStation] = field(default_factory=list[RadioStation])


@dataclass
class StarRecord:
	"""Playtime in seconds, 0 to 10 rating, loved/hated status & timestamp

	Hate status is currently not implemented. Integrations such as ListenBrainz use it.
	"""

	playtime: float = 0
	rating: int = 0
	loved: bool = False
	loved_timestamp: float = 0
	hated: bool = False
	hated_timestamp: float = 0


@dataclass
class TauonQueueItem:
	"""TauonQueueItem is [trackid, position, playlist_id, type, album_stage, uid_gen(), auto_stop]

	type:
		0 is a track
		1 is an album

	Old pre-migration queue[6]-style numbering help table:
		0 track_id    (int)
		1 position    (int)
		2 playlist_id (int)
		3 type        (int)
		4 album_stage (int)
		5 uuid_int    (int)
		6 auto_stop   (bool)
	"""

	track_id: int
	position: int
	playlist_id: int
	type: QueueType
	album_stage: int
	uuid_int: int
	auto_stop: bool


@dataclass
class TauonPlaylist:
	"""Playlist is [Name, playing, playlist_ids, position, hide folder title, selected, uid (1 to 100000000), last_folder, hidden(bool)]

	Old pre-migration pl[6]-style numbering help table:
		0 title (string)
		1 playing (int)
		2 playlist_ids (list of int)
		3 position (int)
		4 hide_title on playlist folders (bool)
		5 selected (int)
		6 uuid_int (int)
		7 last_folder import paths (list[str])
		8 hidden (bool)
		9 locked (bool)
		10 parent_playlist_id <- Filter (int)
		11 persist_time_positioning
	"""

	title: str
	playing: int
	playlist_ids: list[int]
	position: int  # View Position
	hide_title: bool  # hide playlist folder titles (bool)
	selected: int
	uuid_int: int
	last_folder: list[str]  # last folder import paths
	hidden: bool
	locked: bool
	parent_playlist_id: int  # Filter parent playlist id
	persist_time_positioning: bool  # Persist time positioning
	playlist_file: str = ""  # Playlist may be automatically loaded to and from this filepath
	auto_export: bool = False
	auto_import: bool = False
	relative_export: bool = False
	export_type: str = "xspf"
	file_size: int = 0  # if this is different from last time we'll rescan it


def _(m: str) -> str:
	return m


class Timer:
	"""A seconds based timer"""

	def __init__(self, force: float | None = None) -> None:
		self.start: float = 0
		self.end: float = 0
		self.set()
		if force:
			self.force_set(force)

	def set(self) -> None:
		"""Reset"""
		self.start = time.monotonic()

	def hit(self) -> float:
		"""Return time and reset"""
		self.end = time.monotonic()
		elapsed = self.end - self.start
		self.start = time.monotonic()
		return elapsed

	def get(self) -> float:
		"""Return time only"""
		self.end = time.monotonic()
		return self.end - self.start

	def force_set(self, sec: float) -> None:
		self.start = time.monotonic()
		self.start -= sec


class TestTimer:
	"""Simple bool timer object"""

	def __init__(self, time: float) -> None:
		self.timer = Timer()
		self.time = time

	def test(self) -> bool:
		return self.timer.get() > self.time


class ColourGenCache:
	def __init__(self, saturation: float, luminance: float) -> None:
		self.saturation = saturation
		self.luminance = luminance
		self.store: dict[str, ColourRGBA] = {}

	def get(self, key: str) -> ColourRGBA:
		if key in self.store:
			return self.store[key]

		colour = random_colour(self.saturation, self.luminance)

		self.store[key] = colour
		return colour


class FunctionStore:
	"""Stores functions and arguments for calling later"""

	def __init__(self) -> None:
		self.items: list[tuple[Callable[..., None], tuple]] = []

	def store(self, function: Callable[..., None], args: tuple = ()) -> None:
		self.items.append((function, args))

	def recall_all(self) -> None:
		while self.items:
			item = self.items.pop()
			item[0](*item[1])


j_chars = "あおいえうんわらまやはなたさかみりひにちしきるゆむぬつすくれめへねてせけをろもほのとそこアイウエオンヲラマハナタサカミヒニチシキルユムフヌツスクレメヘネテセケロヨモホノトソコ"
year_search = re.compile(r"\d{4}")
# Pre-compile the regular expression pattern for dates like YYYY-MM-DD or DD-MM-YYYY
date_pattern = re.compile(r"\b(?:\d{2}([/. -])\d{2}\1(\d{4})|\b(\d{4})([/. -])\d{2}\4\d{2}).*")
# YYYY or YYYY-MM
year_month_pattern = re.compile(r"^(\d{4})$|^(\d{4})([/. -])\d{2}$")

genre_corrections = [
	"J-Pop",
	"J-Rock",
	"K-Pop",
	"Hip-Hop",
]
genre_corrections2 = [x.lower().replace("-", "").replace(" ", "") for x in genre_corrections]

mac_styles = {
	"mac": None,
	"whitesur": None,
	"vimix": None,
	"sweet": None,
	"dracula": [ColourRGBA(248, 58, 67, 255), ColourRGBA(239, 251, 122, 255), ColourRGBA(74, 254, 104, 255)],
	"nordic": None,
	"juno": None,
}

id3_genre_dict = {
	0: "Blues",
	1: "Classic Rock",
	2: "Country",
	3: "Dance",
	4: "Disco",
	5: "Funk",
	6: "Grunge",
	7: "Hip-Hop",
	8: "Jazz",
	9: "Metal",
	10: "New Age",
	11: "Oldies",
	12: "Other",
	13: "Pop",
	14: "R&B",
	15: "Rap",
	16: "Reggae",
	17: "Rock",
	18: "Techno",
	19: "Industrial",
	20: "Alternative",
	21: "Ska",
	22: "Death Metal",
	23: "Pranks",
	24: "Soundtrack",
	25: "Euro-Techno",
	26: "Ambient",
	27: "Trip-Hop",
	28: "Vocal",
	29: "Jazz+Funk",
	30: "Fusion",
	31: "Trance",
	32: "Classical",
	33: "Instrumental",
	34: "Acid",
	35: "House",
	36: "Game",
	37: "Sound Clip",
	38: "Gospel",
	39: "Noise",
	40: "Alternative Rock",
	41: "Bass",
	42: "Soul",
	43: "Punk",
	44: "Space",
	45: "Meditative",
	46: "Instrumental Pop",
	47: "Instrumental Rock",
	48: "Ethnic",
	49: "Gothic",
	50: "Darkwave",
	51: "Techno-Industrial",
	52: "Electronic",
	53: "Pop-Folk",
	54: "Eurodance",
	55: "Dream",
	56: "Southern Rock",
	57: "Comedy",
	58: "Cult",
	59: "Gangsta Rap",
	60: "Top 40",
	61: "Christian Rap",
	62: "Pop/Funk",
	63: "Jungle",
	64: "Native American",
	65: "Cabaret",
	66: "New Wave",
	67: "Psychedelic",
	68: "Rave",
	69: "Showtunes",
	70: "Trailer",
	71: "Lo-Fi",
	72: "Tribal",
	73: "Acid Punk",
	74: "Acid Jazz",
	75: "Polka",
	76: "Retro",
	77: "Musical",
	78: "Rock & Roll",
	79: "Hard Rock",
	80: "Folk",
	81: "Folk-Rock",
	82: "National Folk",
	83: "Swing",
	84: "Fast Fusion",
	85: "Bebob",
	86: "Latin",
	87: "Revival",
	88: "Celtic",
	89: "Bluegrass",
	90: "Avantgarde",
	91: "Gothic Rock",
	92: "Progressive Rock",
	93: "Psychedelic Rock",
	94: "Symphonic Rock",
	95: "Slow Rock",
	96: "Big Band",
	97: "Chorus",
	98: "Easy Listening",
	99: "Acoustic",
	100: "Humour",
	101: "Speech",
	102: "Chanson",
	103: "Opera",
	104: "Chamber Music",
	105: "Sonata",
	106: "Symphony",
	107: "Booty Bass",
	108: "Primus",
	109: "Porn Groove",
	110: "Satire",
	111: "Slow Jam",
	112: "Club",
	113: "Tango",
	114: "Samba",
	115: "Folklore",
	116: "Ballad",
	117: "Power Ballad",
	118: "Rhythmic Soul",
	119: "Freestyle",
	120: "Duet",
	121: "Punk Rock",
	122: "Drum Solo",
	123: "A Cappella",
	124: "Euro-House",
	125: "Dance Hall",
	126: "Goa",
	127: "Drum & Bass",
	128: "Club-House",
	129: "Hardcore",
	130: "Terror",
	131: "Indie",
	132: "BritPop",
	133: "Negerpunk",
	134: "Polsk Punk",
	135: "Beat",
	136: "Christian Gangsta Rap",
	137: "Heavy Metal",
	138: "Black Metal",
	139: "Crossover",
	140: "Contemporary Christian",
	141: "Christian Rock",
	142: "Merengue",
	143: "Salsa",
	144: "Thrash Metal",
	145: "Anime",
	146: "JPop",
	147: "Synthpop",
	148: "Abstract",
	149: "Art Rock",
	150: "Baroque",
	151: "Bhangra",
	152: "Big Beat",
	153: "Breakbeat",
	154: "Chillout",
	155: "Downtempo",
	156: "Dub",
	157: "EBM",
	158: "Eclectic",
	159: "Electro",
	160: "Electroclash",
	161: "Emo",
	162: "Experimental",
	163: "Garage",
	164: "Global",
	165: "IDM",
	166: "Illbient",
	167: "Industro-Goth",
	168: "Jam Band",
	169: "Krautrock",
	170: "Leftfield",
	171: "Lounge",
	172: "Math Rock",
	173: "New Romantic",
	174: "Nu-Breakz",
	175: "Post-Punk",
	176: "Post-Rock",
	177: "Psytrance",
	178: "Shoegaze",
	179: "Space Rock",
	180: "Trop Rock",
	181: "World Music",
	182: "Neoclassical",
	183: "Audiobook",
	184: "Audio Theatre",
	185: "Neue Deutsche Welle",
	186: "Podcast",
	187: "Indie Rock",
	188: "G-Funk",
	189: "Dubstep",
	190: "Garage Rock",
	191: "Psybient",
	192: "Unknown",
}


def tmp_cache_dir() -> str:
	return str(Path(GLib.get_tmp_dir()) / "TauonMusicBox")


def point_proximity_test(a: list[int], b: list[int], p: float) -> bool:
	"""Test given proximity between two 2d points to given square"""
	return abs(a[0] - b[0]) < p and abs(a[1] - b[1]) < p


def point_distance(a: list[int], b: list[int]) -> float:
	"""Get distance between two points"""
	return math.sqrt(abs(a[0] - b[0]) ** 2 + abs(a[1] - b[1]) ** 2)


def rm_16(line: str) -> str:
	"""Removes whatever this is from a line, I forgot"""
	if "ÿ þ" in line:
		for c in line:
			line = line[1:]
			if c == "þ":
				break

		line = line[::2]
	return line


def get_display_time(seconds: float) -> str:
	"""Returns a string from seconds to a compact time format, e.g 2h:23"""
	if math.isinf(seconds) or math.isnan(seconds):
		logging.error("Infinite/NaN time passed to get_display_time()!")
		return "??:??"
	result = divmod(int(seconds), 60)
	if result[0] > 99:
		result = divmod(result[0], 60)
		return str(result[0]) + "h " + str(result[1]).zfill(2)
	return str(result[0]).zfill(2) + ":" + str(result[1]).zfill(2)


def get_hms_time(seconds: float) -> str:
	m, s = divmod(round(seconds), 60)
	h, m = divmod(m, 60)
	if h:
		return f"{h:d}:{m:02d}:{s:02d}"
	return f"{m:02d}:{s:02d}"


def hms_to_seconds(time_str: str) -> int:
	components = time_str.split(":")
	seconds = 0
	if len(components) == 1:  # If only seconds provided
		seconds = int(components[0])
	elif len(components) == 2:  # If minutes and seconds provided
		seconds = int(components[0]) * 60 + int(components[1])
	elif len(components) == 3:  # If hours, minutes, and seconds provided
		seconds = int(components[0]) * 3600 + int(components[1]) * 60 + int(components[2])
	return seconds


def get_filesize_string(file_bytes: int, rounding: int = 2) -> str:
	"""Creates a string from number of bytes to X MB/kB etc with Locale adjustment"""
	if not file_bytes:
		return "0"
	if file_bytes < 1000:
		line = locale.str(file_bytes) + _(" B")
	elif file_bytes < 1000000:
		file_kb = round(file_bytes / 1000, rounding)
		line = locale.str(file_kb) + _(" KB")
	elif file_bytes < 1000000000:
		file_mb = round(file_bytes / 1000000, rounding)
		line = locale.str(file_mb) + _(" MB")
	else:
		file_mb = round(file_bytes / 1000000000, 1)
		line = locale.str(file_mb) + _(" GB")
	return line


def get_filesize_string_rounded(file_bytes: int) -> str:
	if not file_bytes:
		return "0"
	if file_bytes < 1000:
		line = str(round(file_bytes)) + _(" B")
	elif file_bytes < 1000000:
		file_kb = round(file_bytes / 1000)
		line = str(file_kb) + _(" KB")
	else:
		file_mb = round(file_bytes / 1000000, 1)
		line = str(file_mb) + _(" MB")
	return line


def get_modify_date_string(modify_date: float) -> str:
	ago = round(time.time() - modify_date)
	if ago < 0:
		ago_str = _("Not yet")
	elif ago < 60:  # < minute
		ago_str = _("Now")
	else:
		ago = ago // 60
		if ago < 60:  # minutes
			ago_str = str(ago) + _("min")
		else:
			ago = ago // 60
			if ago < 24:  # hours
				ago_str = str(ago) + _("hr")
			else:
				ago = ago // 24
				if ago < 7:  # days
					ago_str = str(ago) + _("dy")
				elif ago < 31:  # weeks
					ago_str = str(ago // 7) + _("wk")
				elif ago < 90:  # months. dont worry about it
					ago_str = str(ago // 30) + _("mnth")
				else:
					try:
						modify_date = time.localtime(modify_date)
					except Exception:
						ago_str = _("Unknown")
					else:
						if ago < 365:
							ago_str = time.strftime("%d %b", modify_date)
						else:
							ago_str = time.strftime("%b '%y", modify_date)
	return ago_str


def test_lumi(c1: ColourRGBA) -> float:
	"""Estimates the perceived luminance of a colour"""
	return 1 - (0.299 * c1.r + 0.587 * c1.g + 0.114 * c1.b) / 255


def rel_luminance(colour: ColourRGBA) -> float:
	r = colour.r / 255
	g = colour.g / 255
	b = colour.b / 255

	if r < 0.03928:
		r /= 12.90
	else:
		r = ((r + 0.055) / 1.055) ** 2.4

	if g < 0.03928:
		g /= 12.90
	else:
		g = ((g + 0.055) / 1.055) ** 2.4

	if b < 0.03928:
		b /= 12.90
	else:
		b = ((b + 0.055) / 1.055) ** 2.4

	return 0.2126 * r + 0.7152 * g + 0.0722 * b


def contrast_ratio(c1: ColourRGBA, c2: ColourRGBA) -> float:
	l1 = rel_luminance(c1)
	l2 = rel_luminance(c2)

	if l2 > l1:
		l2, l1 = l1, l2

	return (l1 + 0.05) / (l2 + 0.05)


def colour_value(c1: ColourRGBA) -> int:
	"""Give the sum of first 3 elements in a list"""
	return c1.r + c1.g + c1.b


def alpha_blend(colour: ColourRGBA, base: ColourRGBA) -> ColourRGBA:
	"""Performs alpha blending of one colour (RGB-A) onto another (RGB)"""
	alpha = colour.a / 255
	return ColourRGBA(
		int(alpha * colour.r + (1 - alpha) * base.r),
		int(alpha * colour.g + (1 - alpha) * base.g),
		int(alpha * colour.b + (1 - alpha) * base.b),
		255,
	)


def alpha_mod(colour: ColourRGBA, alpha: int) -> ColourRGBA:
	"""Change the alpha component of an RGBA list"""
	return ColourRGBA(colour.r, colour.g, colour.b, alpha)


def colour_slide(a: ColourRGBA, b: ColourRGBA, x: int, x_limit: int) -> ColourRGBA:
	"""Shift between two colours based on x where x is between 0 and limit"""
	return ColourRGBA(
		min(int(a.r + ((b.r - a.r) * (x / x_limit))), 255),
		min(int(a.g + ((b.g - a.g) * (x / x_limit))), 255),
		min(int(a.b + ((b.b - a.b) * (x / x_limit))), 255),
		255,
	)


def hex_to_rgb(colour: str) -> list[int]:
	colour = colour.strip("#")
	return [int(colour[i : i + 2], 16) for i in (0, 2, 4)] + [255]


def check_equal(lst: list[int]) -> bool:
	"""Check if all the numbers in a list are the same"""
	return not lst or lst.count(lst[0]) == len(lst)


def is_grey(lst: ColourRGBA) -> bool:
	"""Check if R, G, and B have the same values"""
	return lst.r == lst.g == lst.b


def star_count(sec: float, dur: float) -> int:
	"""Give a score from 0-7 based on number of seconds"""
	stars = 0
	if dur and sec / dur > 0.95:
		stars += 1
	if sec > 60 * 15:
		stars += 1
	if sec > 60 * 30:
		stars += 1
	if sec > 60 * 60:
		stars += 1
	if sec > 60 * 60 * 2:
		stars += 1
	if sec > 60 * 60 * 4:
		stars += 1
	if sec > 60 * 60 * 8:
		stars += 1
	if sec > 60 * 60 * 16:
		stars += 1
	return stars


def star_count3(sec: float, dur: float) -> int:
	stars = 0
	if dur and sec / dur > 0.95:
		stars += 1
	if dur and sec / dur > 1.95:
		stars += 1
	if sec > 60 * 5:
		stars += 1
	if sec > 60 * 10:
		stars += 1
	if sec > 60 * 15:
		stars += 1
	if sec > 60 * 20:
		stars += 1
	if sec > 60 * 30:
		stars += 1
	if sec > 60 * 60:
		stars += 1
	if sec > 60 * 60 * 1.5:
		stars += 1
	if sec > 60 * 60 * 1.75:
		stars += 1
	# if sec > 60 * 60 * 2:
	# stars += 1
	return stars


def star_count2(sec: float) -> float:
	"""Give a score from 0.0 - 1.0 based on number of seconds"""
	star = 0
	star += min(sec / (60 * 4), 0.2)
	star += min(sec / (60 * 60), 0.4)
	star += sec / (60 * 60 * 10)
	return float(round(min(star, 1), 1))


def search_magic(terms: str, evaluate: str) -> bool:
	return all(word in evaluate for word in terms.split())


def search_magic_any(terms: str, evaluate: str) -> bool:
	return any(word in evaluate for word in terms.split())


def random_colour(saturation: float, luminance: float) -> ColourRGBA:
	h = round(random.random(), 2)
	colour = colorsys.hls_to_rgb(h, luminance, saturation)
	return ColourRGBA(int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255)


def hsl_to_rgb(h: float, s: float, l: float) -> ColourRGBA:
	colour = colorsys.hls_to_rgb(h, l, s)
	return ColourRGBA(int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255)


def hls_to_rgb(h: float, l: float, s: float) -> ColourRGBA:
	"""Duplicate HSL function so it works for the less common alt name too"""
	colour = colorsys.hls_to_rgb(h, l, s)
	return ColourRGBA(int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255)


def rgb_to_hls(r: float, g: float, b: float) -> tuple[float, float, float]:
	return colorsys.rgb_to_hls(r / 255, g / 255, b / 255)


def rgb_add_hls(source: ColourRGBA, h: float = 0, l: float = 0, s: float = 0) -> ColourRGBA:
	c = colorsys.rgb_to_hls(source.r / 255, source.g / 255, source.b / 255)
	colour = colorsys.hls_to_rgb(c[0] + h, min(max(c[1] + l, 0), 1), min(max(c[2] + s, 0), 1))
	return ColourRGBA(int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), source.a)


def is_light(colour: ColourRGBA) -> bool:
	return test_lumi(colour) < 0.2


def folder_file_scan(path: str, extensions: set[str]) -> float:
	match = 0
	count = sum([len(files) for r, d, files in os.walk(path)])
	for ext in extensions:
		match += len(glob.glob(path + "/**/*." + ext.lower(), recursive=True))

	if count == 0:
		return 0

	if count < 5 and match > 0:
		return 1

	return match / count


def is_ignorable_file(string: str) -> bool:
	return any(s in string for s in ["Thumbs.db", ".log", "desktop.ini", "DS_Store", ".nfo", "yric"])


def get_year_from_string(s: str) -> str:
	"""Gets year in form of YYYY from a string

	Example usage:
		example_string = "Event date: 2021-12-31."
		print(get_year_from_string(example_string))
		> "2021"
	"""
	# Search for the pattern in the string
	match = date_pattern.search(s)

	# Extract and return the year if a match is found
	if match:
		return match.group(2) or match.group(3)

	match2 = year_month_pattern.search(s)
	if match2:
		return match2.group(1) or match2.group(2)

	return ""


def is_music_related(string: str) -> bool:
	for s in [
		"Folder.jpg",
		"folder.jpg",
		"Cover.jpg",
		"cover.jpg",
		"AlbumArt",
		".m3u",
		".m3u8",
		".cue",
		".CUE",
	]:
		if s in string:
			return True
	return False


def archive_file_scan(path: str, extensions: set[str], launch_prefix: str = "") -> float:
	"""Get ratio of given file extensions in archive"""
	ext = os.path.splitext(path)[1][1:].lower()
	# logging.info(path)
	# logging.info(ext)
	try:
		if ext == "rar":
			matches = 0
			count = 0
			line = (
				launch_prefix + "unrar lb -p- " + shlex.quote(path) + " " + shlex.quote(os.path.dirname(path)) + os.sep
			)
			result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE, check=True)
			file_list = result.stdout.decode("utf-8", "ignore").split("\n")
			# logging.info(file_list)
			for fi in file_list:
				for ty in extensions:
					if fi[len(ty) * -1 :].lower() == ty:
						matches += 1
						break
					if is_ignorable_file(fi):
						count -= 1
						break
					if is_music_related(fi):
						matches += 5
				count += 1
			if count > 200:
				# logging.info("RAR archive has many files")
				# logging.info("   --- " + path)
				return 0
			if matches == 0:
				# logging.info("RAR archive does not appear to contain audio files")
				# logging.info("   --- " + path)
				return 0
			if count == 0:
				# logging.info("Archive has no files")
				# logging.info("   --- " + path)
				return 0

		elif ext == "7z":
			matches = 0
			count = 0
			line = launch_prefix + "7z l " + shlex.quote(path)  # + " " + shlex.quote(os.path.dirname(path)) + os.sep
			result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE, check=True)
			file_list = result.stdout.decode("utf-8", "ignore").split("\n")
			# logging.info(file_list)

			for fi in file_list:
				if "....A" not in fi:
					continue
				for ty in extensions:
					if fi[len(ty) * -1 :].lower() == ty:
						matches += 1
						break
					if is_ignorable_file(fi):
						count -= 1
						break
					if is_music_related(fi):
						matches += 5
				count += 1

			if count > 200:
				# logging.info("7z archive has many files")
				# logging.info("   --- " + path)
				return 0
			if matches == 0:
				# logging.info("7z archive does not appear to contain audio files")
				# logging.info("   --- " + path)
				return 0
			if count == 0:
				# logging.info("7z archive has no files")
				# logging.info("   --- " + path)
				return 0

		elif ext == "zip":
			with zipfile.ZipFile(path, "r") as zip_ref:
				matches = 0
				count = 0
				# logging.info(zip_ref.namelist())
				for fi in zip_ref.namelist():
					for ty in extensions:
						if fi[len(ty) * -1 :].lower() == ty:
							matches += 1
							break
						if is_ignorable_file(fi):
							count -= 1
							break
						if is_music_related(fi):
							matches += 5
					count += 1
				if count == 0:
					# logging.info("Archive has no files")
					# logging.info("   --- " + path)
					return 0
				if count > 300:
					# logging.info("Zip archive has many files")
					# logging.info("   --- " + path)
					return 0
				if matches == 0:
					# logging.info("Zip archive does not appear to contain audio files")
					# logging.info("   --- " + path)
					return 0
		else:
			return 0

	except Exception:
		logging.exception("Archive test error")

		return 0

	if count == 0:
		return 0

	ratio = matches / count
	if count < 5 and matches > 0:
		ratio = 100
	return ratio


def get_folder_size(path: str) -> int:
	total_size = 0
	for dirpath, dirnames, filenames in os.walk(path):
		for f in filenames:
			fp = os.path.join(dirpath, f)
			total_size += os.path.getsize(fp)
	return total_size


def filename_safe(text: str, sub: str = "") -> str:
	for cha in '/\\<>:"|?*':
		text = text.replace(cha, sub)
	return text.rstrip(" .")


def filename_to_metadata(filename: str) -> tuple[str, str]:
	# Remove the file extension
	name_without_extension = filename.rsplit(".", 1)[0]

	# Remove leading track numbers if present
	name_without_track_number = re.sub(r"^\d{1,2}[. -]+", "", name_without_extension)

	# Split the filename on ' - ' to separate the artist and the title
	if " - " in name_without_track_number:
		artist, title = name_without_track_number.split(" - ", 1)
	else:
		# If ' - ' is not present, return the whole name as title and empty string as artist
		artist = ""
		title = name_without_track_number

	return artist, title


def get_artist_strip_feat(track_object: TrackClass) -> str:
	artist_name = track_object.artist  # .lower()
	if track_object.misc.get("artists"):
		artist_name = track_object.misc["artists"][0]
	elif track_object.album_artist:
		if (
			"feat." in artist_name
			or "pres." in artist_name
			or ", " in artist_name
			or "; " in artist_name
			or "& " in artist_name
			or "ft. " in artist_name
			or not artist_name
		):
			if track_object.album_artist.lower() != "va" and track_object.album_artist.lower() != "various artists":
				artist_name = track_object.album_artist
	return artist_name


def get_artist_safe(track: TrackClass | None) -> str:
	if track:
		artist = track.album_artist
		if not artist:
			artist = track.artist
		artist = filename_safe(artist)
		artist = artist.split("feat")[0]
		artist = artist.split(", ")[0]
		return artist.split("; ")[0]
	return ""


def get_split_artists(track: TrackClass) -> list[str]:
	if "artists" in track.misc:
		return track.misc["artists"]
	artist = track.artist.split("feat")[0].strip()
	return re.split(r"; |, |& ", artist)


def coll_rect(rect1: list[int], rect2: list[int]) -> bool:
	return not (
		rect1[0] + rect1[2] < rect2[0]
		or rect1[1] + rect1[3] < rect2[1]
		or rect1[0] > rect2[0] + rect2[2]
		or rect1[1] > rect2[1] + rect2[3]
	)


def commonprefix(parents: str) -> str:
	cp: list[str] = []
	ls = [p.split("/") for p in parents]
	ml = min(len(p) for p in ls)

	for i in range(ml):
		s = set(p[i] for p in ls)
		if len(s) != 1:
			break

		cp.append(s.pop())

	return "/".join(cp)


def fader_timer(time_point: float, start: float, duration: float, off: bool = True, fade_range: int = 255) -> int:
	if time_point < start:
		fade = fade_range
	elif time_point < start + duration:
		p = (time_point - start) / duration
		fade = int(fade_range - (fade_range * p))
	else:
		fade = 0

	return fade


def grow_rect(rect: tuple[int, int, int, int], px: int) -> tuple[int, int, int, int]:
	return rect[0] - px, rect[1] - px, rect[2] + px * 2, rect[3] + px * 2


# def get_hash(f_path, mode='sha256'):
# h = hashlib.new(mode)
# with open(f_path, 'rb') as file:
# data = file.read()
# h.update(data)
# digest = h.hexdigest()
# return digest


def subtract_rect(
	base: tuple[int, int, int, int],
	hole: tuple[int, int, int, int],
) -> tuple[
	tuple[int, int, int, int],
	tuple[int, int, int, int],
	tuple[int, int, int, int],
	tuple[int, int, int, int],
]:
	"""Return 4 rects from 1 minus 1 inner (with overlaps)"""
	west = base[0], base[1], hole[0], base[3]
	north = base[0], base[1], base[2], hole[1] - base[1]
	east = base[0] + hole[0] + hole[2], base[1], base[2] - (hole[0] + hole[2]), base[3]
	south = base[0], hole[1] + hole[3], base[2], base[3] - hole[3] - 2

	return west, north, east, south


def genre_correct(text: str) -> str:
	parsed = text.lower().replace("-", "").replace(" ", "").strip()
	if parsed.startswith("post"):
		return ("Post-" + parsed[4:]).title()
	if parsed in genre_corrections2:
		return genre_corrections[genre_corrections2.index(parsed)]
	return text.title().strip()


def reduce_paths(paths: list[str]) -> None:
	"""In-place remove of redundant sub-paths from list of folder paths"""
	paths[:] = list(set(paths))[:]  # remove duplicates

	while "" in paths:
		paths.remove("")

	while True:
		remove_path = False
		for i in reversed(range(len(paths))):
			path = paths[i].rstrip("\\/")

			for b in reversed(range(len(paths))):
				path2 = paths[b].rstrip("\\/")

				if len(path) > len(path2) and path.startswith(path2) and path[len(path2)] in ("/", "\\"):
					del paths[i]
					remove_path = True
					break

			if remove_path:
				break
		if not remove_path:
			break


def fit_box(inner: tuple[int, int], outer: tuple[int, int]) -> tuple[int, int]:
	scale = min(outer[0] / inner[0], outer[1] / inner[1])
	return round(inner[0] * scale), round(inner[1] * scale)


def seconds_to_day_hms(seconds: float, s_day: float, s_days: float) -> str:
	days, seconds = divmod(seconds, 86400)
	hours, seconds = divmod(seconds, 3600)
	minutes, seconds = divmod(seconds, 60)

	if days == 1:
		return f"{int(days)!s} {s_day}, {int(hours)!s}:{int(minutes)!s}:{int(seconds)!s}"
	return f"{int(days)!s} {s_days}, {int(hours)!s}:{str(int(minutes)).zfill(2)}:{str(int(seconds)).zfill(2)}"


def shooter(func: Callable[..., None], args: tuple = ()) -> None:
	shoot = threading.Thread(target=func, args=args)
	shoot.daemon = True
	shoot.start()


def d_date_display(track: TrackClass) -> str:
	if "rdat" in track.misc and str(track.date) != track.misc["rdat"]:
		return str(track.date) + " → " + track.misc["rdat"]
	return str(track.date)


def d_date_display2(track: TrackClass) -> str:
	if "rdat" in track.misc and str(get_year_from_string(track.date)) != get_year_from_string(track.misc["rdat"]):
		return str(get_year_from_string(track.date)) + " → " + get_year_from_string(track.misc["rdat"])
	return str(get_year_from_string(track.date))


def process_odat(nt: TrackFile, odat: str) -> None:
	if (
		odat
		and odat != nt.date
		and odat != nt.date[:4]
		and odat != nt.date[-4:]
		and nt.date != odat[:4]
		and nt.date != odat[-4:]
	):
		if not nt.date:
			nt.date = odat
		else:
			nt.misc["rdat"] = nt.date
			nt.date = odat


def clean_string(s: str) -> str:
	return s.encode("utf-8", "surrogatepass").decode("utf-8", "replace")


def uri_parse(s: str) -> str:
	if s.startswith("file://"):
		s = str(urllib.parse.unquote(s)).replace("file://", "").replace("\r", "")
	return s


def sleep_timeout(condition_function: Callable[[], bool], time_limit: int = 2) -> None:
	if condition_function():
		return
	timer = Timer()
	timer.set()
	while not condition_function():
		time.sleep(0.01)
		if timer.get() > time_limit:
			break


def tryint(string: str) -> int | str:
	try:
		return int(string)
	except ValueError:
		return string
	except Exception:
		logging.exception("Unknown error trying to convert string to int!")
		return string


class FPSCounter:
	"""A robust FPS counter that handles pauses and provides smooth measurements.

	Args:
		window_size: Number of frames to average over (default: 60)
		min_update_interval: Minimum time between updates in seconds (default: 0.1)
		max_frame_time: Maximum expected frame time in seconds. If exceeded,
						frame is considered anomalous (default: 1.0)

	"""

	def __init__(self, window_size: int = 60, min_update_interval: float = 0.1, max_frame_time: float = 1.0) -> None:
		self.window_size = window_size
		self.min_update_interval = min_update_interval
		self.max_frame_time = max_frame_time

		self.frame_times = deque(maxlen=window_size)
		self.last_tick_time: float | None = None
		self.last_update_time: float | None = None
		self.cached_fps: float = 0.0
		self.frame_count: int = 0

	def tick(self) -> None:
		"""Call this once per frame to record the frame timing."""
		current_time = time.perf_counter()

		if self.last_tick_time is not None:
			frame_time = current_time - self.last_tick_time

			# Ignore anomalous frame times (e.g., after a pause)
			if frame_time <= self.max_frame_time:
				self.frame_times.append(frame_time)

		self.last_tick_time = current_time
		self.frame_count += 1

	def get(self) -> float:
		"""Get the current FPS value.

		Returns the cached value if called within min_update_interval,
		otherwise recalculates based on recent frame times.

		Returns:
			float: Current FPS value, or 0.0 if insufficient data

		"""
		current_time = time.perf_counter()

		# Return cached value if we updated recently
		if self.last_update_time is not None and current_time - self.last_update_time < self.min_update_interval:
			return self.cached_fps

		# Calculate new FPS value
		if len(self.frame_times) == 0:
			self.cached_fps = 0.0
		else:
			avg_frame_time = sum(self.frame_times) / len(self.frame_times)
			self.cached_fps = 1.0 / avg_frame_time if avg_frame_time > 0 else 0.0

		self.last_update_time = current_time
		return self.cached_fps

	def reset(self) -> None:
		"""Reset the FPS counter."""
		self.frame_times.clear()
		self.last_tick_time = None
		self.last_update_time = None
		self.cached_fps = 0.0
		self.frame_count = 0



from __future__ import annotations

import json
import logging
import subprocess
import threading
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
	from collections.abc import Callable
	from pathlib import Path


class MacNowPlayingHelper:
	"""Spawn and communicate with the macOS Now Playing helper app.

	IPC is newline-delimited JSON over stdin/stdout.
	"""

	def __init__(
		self,
		executable: Path,
		on_media_key: Callable[[str], None],
		on_seek: Callable[[float], None] | None = None,
		on_seek_relative: Callable[[float], None] | None = None,
	) -> None:
		self.executable = executable
		self._on_media_key = on_media_key
		self._on_seek = on_seek
		self._on_seek_relative = on_seek_relative
		self._proc: subprocess.Popen[str] | None = None
		self._stdin_lock = threading.Lock()
		self._reader_thread: threading.Thread | None = None

	def start(self) -> bool:
		if self._proc is not None:
			return True
		try:
			self._proc = subprocess.Popen(
				[str(self.executable)],
				stdin=subprocess.PIPE,
				stdout=subprocess.PIPE,
				stderr=subprocess.DEVNULL,
				text=True,
				bufsize=1,
			)
		except Exception:
			logging.exception("Failed to start macOS Now Playing helper")
			self._proc = None
			return False

		self._reader_thread = threading.Thread(target=self._read_loop, name="TauonNowPlaying", daemon=True)
		self._reader_thread.start()
		return True

	def _read_loop(self) -> None:
		assert self._proc is not None
		assert self._proc.stdout is not None
		for line in self._proc.stdout:
			line = line.strip()
			if not line:
				continue
			try:
				msg = json.loads(line)
			except Exception:
				logging.debug(f"NowPlaying helper sent non-JSON line: {line!r}")
				continue

			if not isinstance(msg, dict):
				continue

			mtype = msg.get("type")
			if mtype == "media_key":
				name = msg.get("name")
				if isinstance(name, str):
					try:
						self._on_media_key(name)
					except Exception:
						logging.exception("Error handling media_key from NowPlaying helper")
			elif mtype == "seek":
				pos = msg.get("position")
				if isinstance(pos, (int, float)) and self._on_seek is not None:
					try:
						self._on_seek(float(pos))
					except Exception:
						logging.exception("Error handling seek from NowPlaying helper")
			elif mtype == "seek_relative":
				delta = msg.get("delta")
				if isinstance(delta, (int, float)) and self._on_seek_relative is not None:
					try:
						self._on_seek_relative(float(delta))
					except Exception:
						logging.exception("Error handling seek_relative from NowPlaying helper")

	def send(self, msg: dict[str, Any]) -> None:
		if self._proc is None or self._proc.stdin is None:
			return
		try:
			data = json.dumps(msg, ensure_ascii=False)
		except Exception:
			return
		with self._stdin_lock:
			try:
				self._proc.stdin.write(data + "\n")
				self._proc.stdin.flush()
			except Exception:
				# Helper may have exited.
				return

	def update(
		self,
		*,
		title: str,
		artist: str,
		album: str = "",
		art_path: str = "",
		state: int,
		duration: float | None = None,
		elapsed: float | None = None,
	) -> None:
		msg: dict[str, Any] = {
			"type": "update",
			"title": title,
			"artist": artist,
			"album": album,
			"art_path": art_path,
			"state": state,
		}
		if duration is not None:
			msg["duration"] = float(duration)
		if elapsed is not None:
			msg["elapsed"] = float(elapsed)
		self.send(msg)

	def clear(self) -> None:
		self.send({"type": "clear"})

	def stop(self) -> None:
		if self._proc is None:
			return
		try:
			self.send({"type": "quit"})
		except Exception:
			pass
		try:
			self._proc.terminate()
		except Exception:
			pass
		try:
			self._proc.wait(timeout=1.0)
		except Exception:
			pass
		self._proc = None



"""Tauon Music Box - Lyrics scrape/fetch module

You can add lyric providers in this module

Create a function that takes artist and title, and returns lyrics as a str
If failed to find lyrics, you can return None or raise an exception.
Finally add provider name and function reference to lyric_sources dict below
"""

# Copyright © 2018-2023, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import urllib.parse
from http import HTTPStatus

import requests
from bs4 import BeautifulSoup
from unidecode import unidecode


def ovh(
	artist: str, title: str, user_agent: str = "unused here but it needs to exist in every lyrics function"
) -> tuple[str, str]:
	"""Get lyrics from lyrics.ovh API"""
	q = urllib.parse.quote(f"{artist}/{title}")
	point = f"https://api.lyrics.ovh/v1/{q}"
	r = requests.get(point, timeout=10)
	if r.status_code == HTTPStatus.OK:
		j = r.json()
		return j["lyrics"], ""
	return "", ""


def genius(
	artist: str,
	title: str,
	return_url: bool = False,
	user_agent: str = "unused here but it needs to exist in every lyrics function",
) -> tuple[str, str] | str:
	"""Scrape lyrics from genius.com"""
	artist = artist.split("feat.", maxsplit=1)[0].strip()
	title = title.split("(feat.", maxsplit=1)[0].strip()
	line = f"{artist}-{title}"
	line = re.sub("[,._@!#%^*+:;'()]", "", line)
	line = line.replace("]", "")
	line = line.replace("[", "")
	line = line.replace("?", "")
	line = line.replace(" ", "-")
	line = line.replace("/", "-")
	line = line.replace("-&-", "-and-")
	line = line.replace("&", "-and-")
	line = unidecode(line)
	line = urllib.parse.quote(line)
	line = f"https://genius.com/{line}-lyrics"

	if return_url:
		return line

	page = requests.get(line, timeout=10)
	html = BeautifulSoup(page.text, "html.parser")

	result = html.find("div", class_="lyrics")
	if result is not None:
		lyrics = result.get_text()
		lyrics2: list[str] = []
		for line in lyrics.splitlines():
			if line.startswith("["):
				pass
			else:
				lyrics2.append(line)

		lyrics = "\n".join(lyrics2)
		lyrics = lyrics.strip("\n")
		return lyrics, ""

	results = html.find_all("div", {"class": lambda l: l and "Lyrics__Container" in l})
	if not results:
		return "", ""

	lyrics = "".join([r.get_text("\n") for r in results])

	level = 0
	new = ""
	for cha in lyrics:
		if level <= 0:
			new += cha
		if cha == "[":
			level += 1
		if cha == "]":
			level -= 1
	lyrics = new

	lines = lyrics.splitlines()
	new_lines: list[str] = []
	# fmt:off
	skip_patterns = [
		r"^\d+\s+Contributors?$",   # "3 Contributors"
		r"^[^a-zA-Z]*Lyrics$",      # "[Song Name] Lyrics"
		r"^\s*$",                   # Empty lines
		r"^Embed$",                 # "Embed" button text
		r"^\d+$",                   # Standalone numbers
		r"^See .* Live$",           # "See Artist Live"
		r"^Get tickets as low as",  # Ticket advertisements
		r"^You might also like",    # Related content
		r"^\w+ on Apple Music",     # Apple Music links
	]
	# fmt:on

	for line in lines:
		line = line.strip()

		if not line:
			continue

		should_skip = False
		for pattern in skip_patterns:
			if re.match(pattern, line, re.IGNORECASE):
				should_skip = True
				break

		if should_skip:
			continue

		if "[" in line:
			line = line.split("[", 1)[0].strip()
			if not line:
				continue

		if line.lower() in ["lyrics", "embed", "more on genius"] or line.isdigit() or len(line) < 2:
			continue

		new_lines.append(line.rstrip() + "\n")

	lyrics = "".join(new_lines)
	lyrics = lyrics.replace("(\n", "(")
	lyrics = lyrics.replace("\n)", ")")
	lyrics = lyrics.lstrip("\n")
	lyrics = lyrics.lstrip()

	lyrics_lines = lyrics.split("\n")

	while lyrics_lines:
		first_line = lyrics_lines[0].strip()
		if (
			re.match(r"^\d+\s+Contributors?$", first_line, re.IGNORECASE)
			or re.match(r"^.*Lyrics$", first_line, re.IGNORECASE)
			or first_line.lower() in ["embed", "lyrics"]
			or len(first_line) < 3
		):
			lyrics_lines.pop(0)
		else:
			break

	while lyrics_lines:
		last_line = lyrics_lines[-1].strip()
		if last_line.lower() in ["embed", "more on genius"] or len(last_line) < 3:
			lyrics_lines.pop()
		else:
			break

	final_lyrics = "\n".join(lyrics_lines).strip()

	if len(final_lyrics) > 10:
		return final_lyrics, ""

	return "", ""


def lrclib(artist: str, title: str, user_agent: str = "TauonMusicBox/Devel") -> tuple[str, str]:
	h = {
		"User-Agent": user_agent,
	}

	p = {
		"track_name": title,
		"artist_name": artist,
	}

	r = requests.get("https://lrclib.net/api/get", headers=h, params=p, timeout=10)
	if r.status_code == HTTPStatus.OK:
		p = r.json().get("plainLyrics")
		s = r.json().get("syncedLyrics")
		if p or s:
			return p, s
	return "", ""


def get_lrclib_challenge(user_agent: str = "TauonMusicBox/Devel") -> tuple[str, str]:
	h = {
		"User-Agent": user_agent,
	}
	try:
		r = requests.post("https://lrclib.net/api/request-challenge", headers=h, timeout=10)
	except requests.exceptions.ConnectionError:
		return "", ""
	except Exception:
		logging.exception("Unknown exception")
		return "", ""
	if r.status_code == HTTPStatus.OK:
		p = r.json().get("prefix")
		t = r.json().get("target")
		if p or t:
			return p, t
	return "", ""


lyric_sources = {
	"Genius": genius,
	"lyrics.ovh": ovh,
	"LRCLIB": lrclib,
}

uses_scraping = {
	"Genius",
}



"""Tauon Music Box - Spotify module"""

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import io
import logging
import os
import subprocess
import time
import webbrowser
from typing import TYPE_CHECKING

import requests

from tauon.t_modules.t_draw import QuickThumbnail
from tauon.t_modules.t_extra import Timer

if TYPE_CHECKING:
	from tekore._auth.refreshing import RefreshingCredentials
	from tekore._client.full import Spotify
	from tekore._model.album.full import FullAlbum

	from tauon.t_modules.t_main import Tauon, TrackClass

tekore_imported = False
try:
	import tekore as tk
except ModuleNotFoundError:
	logging.warning("Unable to import Tekore, Spotify support will be disabled.")
except Exception:
	logging.exception("Unknown error trying to import Tekore, Spotify support will be disabled.")
else:
	tekore_imported = True


class SpotCtl:
	def __init__(self, tauon: Tauon) -> None:
		# fmt:off
		self.tauon                   = tauon
		self.strings                 = tauon.strings
		self.show_message            = tauon.show_message
		self.start_timer             = Timer()
		self.status:             int = 0
		self.spotify: Spotify | None = None
		self.loaded_art:         str = ""
		self.playing:           bool = False
		self.coasting:          bool = False
		self.paused:            bool = False
		self.token = None
		self.cred: RefreshingCredentials | None = None
		self.started_once:                 bool = False
		# https://developer.spotify.com/documentation/web-api/concepts/redirect_uri
		self.redirect_uri:                  str = "http://127.0.0.1:7811/spotredir"
		self.current_imports:              dict = {}
		self.spotify_com:                  bool = False
		self.sender = None
		self.cache_saved_albums:      list[str] = []
		self.scope:              str = "user-read-private user-read-playback-position streaming user-modify-playback-state user-library-modify user-library-read user-read-currently-playing user-read-playback-state playlist-read-private playlist-modify-private playlist-modify-public"
		self.launching_spotify: bool = False
		self.preparing_spotify: bool = False
		self.progress_timer:   Timer = Timer()
		self.update_timer:     Timer = Timer()
		# fmt:on

		self.token_path = self.tauon.user_directory / "spot-token-pkce"
		self.pkce_code = None
		self.local: bool = True
		self.country = None

		self.coast_context: str = ""

	def prep_cred(self) -> RefreshingCredentials:
		if not tekore_imported:
			self.show_message(_("Dependency Tekore not found"), mode="warning")
		rc = tk.RefreshingCredentials
		secret = self.tauon.prefs.spot_secret
		if not secret or len(self.tauon.prefs.spot_secret) != 32:
			secret = None
		return rc(client_id=self.tauon.prefs.spot_client, client_secret=secret, redirect_uri=self.redirect_uri)

	def connect(self) -> None:
		if not self.tauon.prefs.spotify_token or not self.tauon.prefs.spot_mode:
			return
		if len(self.tauon.prefs.spot_client) != 32:
			return
		if self.cred is None:
			self.cred = self.prep_cred()
		if self.spotify is None:
			if self.token is None:
				self.load_token()
			if self.token:
				logging.info("Init spotify support")
				self.sender = tk.RetryingSender(retries=3)
				self.spotify = tk.Spotify(self.token, sender=self.sender)
				self.country = self.spotify.current_user().country

	def paste_code(self, code: str) -> None:
		if self.cred is None:
			self.cred = self.prep_cred()

		if self.pkce_code:
			self.token = self.cred.request_pkce_token(code.strip().strip("\n"), self.pkce_code)
		else:
			self.token = self.cred.request_user_token(code.strip().strip("\n"))

		if self.token:
			self.save_token()
			self.show_message(self.strings.spotify_account_connected, mode="done")
		self.tauon.prefs.spot_mode = True

	def save_token(self) -> None:
		if self.token:
			f = self.token_path.open("w", encoding="utf-8")
			f.write(str(self.token.refresh_token))
			f.close()
			self.tauon.prefs.spotify_token = str(self.token.refresh_token)

	def load_token(self) -> None:
		if self.token_path.is_file():
			f = self.token_path.open()
			self.tauon.prefs.spotify_token = f.read().replace("\n", "").strip()
			f.close()

		if self.tauon.prefs.spotify_token:
			secret = self.tauon.prefs.spot_secret
			if not secret or len(self.tauon.prefs.spot_secret) != 32:
				secret = None

			try:
				if secret:
					self.token = tk.refresh_user_token(
						self.tauon.prefs.spot_client, secret, self.tauon.prefs.spotify_token
					)
				else:
					self.token = tk.refresh_pkce_token(self.tauon.prefs.spot_client, self.tauon.prefs.spotify_token)
			except Exception:
				logging.exception("FAILED TO LOAD TOKEN")
				self.show_message(_("Please re-authenticate Spotify in settings"))
				self.tauon.prefs.spotify_token = ""
		else:
			self.show_message(_("Please authenticate Spotify in settings"))

	def delete_token(self) -> None:
		self.tauon.prefs.spotify_token = ""
		self.token = None
		if self.token_path.exists():
			self.token_path.unlink()

	def auth(self) -> None:
		if not tekore_imported:
			self.show_message(
				_("python-tekore not installed"),
				_("If you installed via AUR, you'll need to install this optional dependency, then restart Tauon."),
				mode="error",
			)
			return
		if len(self.tauon.prefs.spot_client) != 32:
			self.show_message(_("Invalid client ID. See Spotify tab in settings."), mode="error")
			return
		if self.cred is None:
			self.cred = self.prep_cred()

		secret = self.tauon.prefs.spot_secret
		if not secret or len(self.tauon.prefs.spot_secret) != 32:
			secret = None

		if not secret:
			logging.info("Using PKCE auth")
			url, self.pkce_code = self.cred.pkce_user_authorisation(scope=self.scope)
		else:
			logging.info("Using user auth")
			self.pkce_code = None
			url = self.cred.user_authorisation_url(scope=self.scope)
		webbrowser.open(url, new=2, autoraise=True)

	def control(self, command: str, param: str | int | None = None) -> None:
		if not self.spotify:
			return

		try:
			if command == "pause" and (self.playing or self.coasting) and not self.paused:
				self.spotify.playback_pause()
				self.paused = True
				self.start_timer.set()
			elif command == "stop" and (self.playing or self.coasting):
				self.paused = False
				self.playing = False
				self.coasting = False
				self.spotify.playback_pause()
				self.start_timer.set()
			elif command == "resume" and (self.coasting or self.playing) and self.paused:
				self.spotify.playback_resume()
				self.paused = False
				self.progress_timer.set()
				self.start_timer.set()
			elif param is None:
				logging.error(f"Passed a command {command} requiring a parameter but did not pass the parameter!")
				return
			elif command == "volume" and type(param) is int:
				self.spotify.playback_volume(param)
			elif command == "seek" and type(param) is int:
				self.spotify.playback_seek(param)
				self.start_timer.set()
			elif command == "next" and type(param) is str:
				self.spotify.playback_next(param)
				# self.start_timer.set()
			elif command == "previous" and type(param) is str:
				self.spotify.playback_previous(param)
				# self.start_timer.set()
			else:
				logging.error(f"Passed an invalid command {command}!")
				return

		except Exception as e:
			logging.exception("Control failure")
			# logging.info(repr(e))
			if "No active device found" in repr(e):
				try:
					tr = self.tauon.pctl.playing_object()
					if command == "resume" and tr and tr.file_ext == "SPTY" and tr.url_key:
						self.show_message("Resuming Spotify playback")
						p = self.tauon.pctl.playing_time
						self.play_target(tr.url_key, p)
						self.tauon.gui.message_box = False
						self.tauon.gui.update += 1
						return
				except Exception:
					logging.exception("Control failure handling failure")
				self.show_message(_("It looks like there are no more active Spotify devices"))

	def get_username(self) -> str | None:
		self.connect()
		if not self.spotify:
			return None
		c = self.spotify.current_user()
		if c and c.id:
			return c.id
		return None

	def add_album_to_library(self, url: str) -> None:
		self.connect()
		if not self.spotify:
			return

		id = url.strip("/").split("/")[-1]

		try:
			self.spotify.saved_albums_add([id])
			if url not in self.cache_saved_albums:
				self.cache_saved_albums.append(url)
		except Exception:
			logging.exception("Error saving album")

	def remove_album_from_library(self, url: str) -> None:
		self.connect()
		if not self.spotify:
			return
		id = url.strip("/").split("/")[-1]

		try:
			self.spotify.saved_albums_delete([id])
			if url in self.cache_saved_albums:
				self.cache_saved_albums.remove(url)
		except Exception:
			logging.exception("Error removing album")

	def get_album_url_from_local(self, track_object: TrackClass) -> str | None:
		if "spotify-album-url" in track_object.misc:
			return track_object.misc["spotify-album-url"]

		self.connect()
		if not self.spotify:
			return None

		results = self.spotify.search(track_object.artist + " " + track_object.album, types=("album",), limit=1)
		for album in results[0].items:
			if "spotify" in album.external_urls:
				return album.external_urls["spotify"]

		return None

	def get_artist_url_from_local(self, track_object: TrackClass) -> str | None:
		if "spotify-artist-url" in track_object.misc:
			return track_object.misc["spotify-artist-url"]

		self.connect()
		if not self.spotify:
			return None

		results = self.spotify.search(track_object.artist, types=("artist",), limit=1)  # + " " + track_object.album
		for artist in results[0].items:
			if "spotify" in artist.external_urls:
				if artist.name.lower() not in track_object.artist.lower():
					return None
				return artist.external_urls["spotify"]

		return None

	def import_all_playlists(self) -> None:
		self.spotify_com = True

		playlists = self.get_playlist_list()
		if playlists:
			for item in playlists:
				self.playlist(item[1], silent=True)
				self.tauon.gui.update += 1
				time.sleep(0.1)

		self.spotify_com = False
		if not playlists:
			self.show_message(self.strings.spotify_need_enable)
			return
		self.show_message(self.strings.spotify_import_complete, mode="done")

	def get_playlist_list(self) -> list[tuple[str, str]] | None:
		self.connect()
		if not self.spotify:
			self.show_message(self.strings.spotify_need_enable)
			return None

		playlists: list[tuple[str, str]] = []
		results = self.spotify.playlists(self.spotify.current_user().id)
		pages = self.spotify.all_pages(results)
		for page in pages:
			items = page.items
			for item in items:
				# if item is None:
				# 	logging.debug("Playlist item is None?")
				# 	continue
				name = item.name
				url = item.external_urls["spotify"]
				playlists.append((name, url))

		return playlists

	def search(self, text: str) -> list[TrackClass] | None:
		self.connect()
		if not self.spotify:
			return None

		results = self.spotify.search(
			text,
			types=("artist", "album", "track"),
			limit=20,
		)

		finds = []

		self.tauon.quickthumbnail.queue.clear()

		if results[0]:
			for i, album in enumerate(results[0].items[1:]):
				if hasattr(album, "album"):
					album = album.album

				img = QuickThumbnail(self.tauon)
				img.url = album.images[-1].url
				img.size = round(50 * self.tauon.gui.scale)
				self.tauon.quickthumbnail.items.append(img)
				if i < 10:
					self.tauon.quickthumbnail.queue.append(img)
				try:
					self.tauon.gall_ren.lock.release()
				except Exception:
					logging.exception("Failed releasing lock!")

				finds.append((11, (album.name, album.artists[0].name), album.external_urls["spotify"], 0, 0, img))

			for artist in results[1].items[0:1]:
				finds.insert(1, (10, artist.name, artist.external_urls["spotify"], 0, 0, None))
			for artist in results[1].items[1:2]:
				finds.insert(11, (10, artist.name, artist.external_urls["spotify"], 0, 0, None))

			for track in results[2].items[0:1]:
				finds.insert(2, (12, (track.name, track.artists[0].name), track.external_urls["spotify"], 0, 0, None))
			for track in results[2].items[5:6]:
				finds.insert(8, (12, (track.name, track.artists[0].name), track.external_urls["spotify"], 0, 0, None))

		return finds

	def search_track(self, track: TrackClass | None) -> None:
		"""Search track on Spotify - returning results is unimplemented"""
		if track is None:
			return

		self.connect()
		if not self.spotify:
			return

		if track.artist and track.title:
			results = self.spotify.search(
				track.artist + " " + track.title,
				types=("track",),
				limit=1,
			)

	def prime_device(self) -> bool | int | None:
		self.connect()
		if not self.spotify:
			return None

		logging.info("Get devices...")
		devices = self.spotify.playback_devices()
		logging.info("Devices found:")
		for device in devices:
			logging.info(f" -- Device name: {device.name}, type: {device.type}, is active: {device.is_active}")

		if devices:
			pass
		else:
			logging.warning("No spotify devices found")

		if not devices:
			return False
		for d in devices:
			if d.is_active:
				return None
		for d in devices:
			if not d.is_restricted:
				return d.id
		return None

	def transfer_to_tauon(self, wait: int = 7) -> None:
		self.preparing_spotify = True
		self.connect()
		if not self.spotify:
			self.preparing_spotify = False
			return

		time.sleep(wait)
		p = self.spotify.playback()
		if not p or not p.is_playing:
			self.show_message(_("Nothing playing"))
			self.preparing_spotify = False
			return

		devices = self.spotify.playback_devices()
		for d in devices:
			if d.name == "Tauon Music Box":
				self.spotify.playback_transfer(d.id, True)
				logging.info("Found Tauon Spotify player")
				break
		else:
			self.show_message(_("Error - Tauon device not found"))

		self.preparing_spotify = False

	def play_target(
		self, id: int | str, force_new_device: bool = False, start_time: int | None = 0, start_callback=None
	) -> None:
		start_time = None if not start_time else int(start_time * 1000)

		self.coasting = False
		self.connect()
		if not self.spotify:
			self.preparing_spotify = False
			self.show_message(
				_("Error. You may need to click Authorise in Settings > Accounts > Spotify."), mode="warning"
			)
			return

		logging.info(f"Want play spotify target {id!s}")
		# Sort devices
		start_new_device = False
		done = False

		logging.info("Get devices...")
		devices = self.spotify.playback_devices()
		logging.info("Devices found:")
		if devices:
			for device in devices:
				logging.info(f" -- Device name: {device.name}, type: {device.type}, is active: {device.is_active}")

		d_id = None
		if not devices or force_new_device:
			logging.info("No devices found...")
			start_new_device = True
		else:
			for device in devices:
				if not device.is_restricted and device.is_active:
					logging.info("Choosing the active device: " + device.name)
					d_id = device.id
					break
			else:
				for device in devices:
					if not device.is_restricted:
						logging.info("Found a device, but its not active...")
						logging.info("Try start track on: " + device.name)
						d_id = device.id

						self.spotify.playback_transfer(device.id)
						time.sleep(0.2)
						self.spotify.playback_start_tracks([id], device_id=d_id, position_ms=start_time)
						tries = 0
						while True:
							result = self.spotify.playback_currently_playing()
							if result and result.is_playing:
								done = True
								self.progress_timer.set()
								logging.info("Looks like starting on that device worked...")
								break
							time.sleep(1)
							tries += 1
							logging.info("Not playing on that inactive device yet, waiting...")
							if tries > 5:
								break

					if d_id or start_new_device or done:
						break

		if not d_id and not start_new_device:
			logging.info("Internal logic error, aborting.")
			return

		# if self.tauon.pctl.playing_state == 1 and self.playing and self.tauon.pctl.playing_time
		# try:
		if start_new_device:
			# if not force_new_device:
			logging.info("Launch new device...")

			self.tauon.gui.update += 1

			logging.info("Initiate device...")
			if self.tauon.prefs.launch_spotify_web:
				logging.info("Open web player...")
				self.launching_spotify = True
				webbrowser.open("https://open.spotify.com/", new=2, autoraise=False)
				tries = 0
				while True:
					time.sleep(2)
					if tries == 0:
						self.tauon.focus_window()
					devices = self.spotify.playback_devices()
					if devices:
						self.progress_timer.set()
						self.spotify.playback_start_tracks([id], device_id=devices[0].id, position_ms=start_time)
						break
					tries += 1
					if tries > 13:
						self.tauon.pctl.stop()
						self.show_message(self.strings.spotify_error_starting, mode="error")
						self.launching_spotify = False
						self.preparing_spotify = False
						self.tauon.gui.update += 1
						return
			else:
				if self.tauon.prefs.launch_spotify_local:
					self.preparing_spotify = True
					logging.info("Queue start librespot...")
					self.tauon.thread_manager.ready_playback()
					self.tauon.pctl.playerCommand = "spotcon"
					self.tauon.pctl.playerCommandReady = True
					if start_callback:
						logging.info("Callback start librespot")
						start_callback()

					# self.tauon.pctl.playing_state = 3
					elif self.tauon.windows:
						self.launching_spotify = True
						appdata = os.getenv("APPDATA")
						if not appdata:
							self.launching_spotify = False
							self.preparing_spotify = False
							logging.info("APPDATA is not set, cannot locate Spotify app.")
							return
						p = appdata + "\\Spotify\\Spotify.exe"
						if not os.path.isfile(p):
							self.launching_spotify = False
							self.preparing_spotify = False
							logging.info("Spotify app exe not found, aborting.")
							return
						subprocess.Popen([p])
					logging.info("Launching spotify app exe")
					time.sleep(3)
				else:
					self.launching_spotify = True
					subprocess.run(["xdg-open", "spotify:track"], check=True)
					time.sleep(3)
					logging.info("Launched spotify app via URI")

				time.sleep(0.5)
				tries = 0

				while True:
					logging.info("Waiting for device ready... try: " + str(tries + 1))
					devices = self.spotify.playback_devices()
					if devices and tries < 13:
						logging.info("Devices found:")
						for device in devices:
							logging.info(
								f" -- Device name: {device.name}, type: {device.type}, is active: {device.is_active}"
							)

						if not self.tauon.prefs.launch_spotify_local:
							self.tauon.focus_window()
						time.sleep(0.5)
						logging.info(f"Selecting device: {devices[0].name}")

						logging.info("Attempt start track...")
						try:
							self.spotify.playback_start_tracks([id], device_id=devices[0].id, position_ms=start_time)
						except Exception:
							logging.exception("Failed to start playback!")
							time.sleep(1)
							tries += 2
							continue
						while True:
							result = self.spotify.playback_currently_playing()
							if result and result.is_playing:
								done = True
								self.progress_timer.set()
								logging.info("Looks like its playing now")
								break
							time.sleep(1)
							tries += 1
							logging.info("Not playing yet, waiting...")
							if tries > 13:
								break

					if done:
						break
					tries += 1
					if tries > 13:
						logging.info("Too many retries, aborting.")
						self.tauon.pctl.stop()
						self.show_message(self.strings.spotify_error_starting, mode="error")
						self.launching_spotify = False
						self.preparing_spotify = False
						self.tauon.gui.update += 1
						return
					time.sleep(1)

			self.launching_spotify = False
			self.preparing_spotify = False
			self.tauon.gui.update += 1

		elif not done:
			# logging.info(d_id)
			logging.info("A ready device is present...")
			try:
				self.progress_timer.set()
				okay = False

				# Check conditions for a proper transition
				if self.playing:
					# logging.info("already playing")
					result = self.spotify.playback_currently_playing()
					if result and result.item and result.is_playing:
						remain = result.item.duration_ms - result.progress_ms
						if 1400 < remain < 3500:
							logging.info("We are close to the end of the current track, queuing next.")
							self.spotify.playback_queue_add("spotify:track:" + id, device_id=d_id)
							okay = True
							logging.info("Waiting for remainder of track...")
							time.sleep(remain / 1000)
							self.progress_timer.set()
							time.sleep(2)
							result = self.spotify.playback_currently_playing()
							if not (result and result.item and result.is_playing):
								logging.info("The queue transition failed")
								okay = False

				# Force a transition
				if not okay:
					logging.info("Starting track on live device.")
					self.spotify.playback_start_tracks([id], device_id=d_id, position_ms=start_time)

			# except tk.client.decor.error.InternalServerError:
			# self.show_message("Spotify server error. Maybe try again later.")
			# return
			except Exception as e:
				logging.exception("Start track on device failed, aborting")
				self.launching_spotify = False
				self.preparing_spotify = False
				self.show_message("Spotify error, try again?", str(e), mode="warning")
				return

		# except Exception as e:
		# logging.exception("Failure. Do you have playback started somewhere?")
		# self.show_message("Error. Do you have playback started somewhere?", mode="error")
		logging.info("Done")
		self.playing = True
		self.started_once = True
		self.launching_spotify = False
		self.preparing_spotify = False
		self.start_timer.set()
		self.tauon.pctl.playing_time = 0
		self.tauon.pctl.decode_time = 0
		self.tauon.gui.pl_update += 1
		self.tauon.thread_manager.ready_playback()

	def get_library_albums(self, return_list: bool = False) -> list | None:
		self.connect()
		if not self.spotify:
			self.spotify_com = False
			self.show_message(self.strings.spotify_need_enable)
			return []

		albums = self.spotify.saved_albums()

		playlist = []
		self.update_existing_import_list()
		self.cache_saved_albums.clear()

		pages = self.spotify.all_pages(albums)

		for page in pages:
			for a in page.items:
				self.load_album(a.album, playlist)

				if a.album.external_urls["spotify"] not in self.cache_saved_albums:
					self.cache_saved_albums.append(a.album.external_urls["spotify"])

		if return_list:
			return playlist

		self.tauon.pctl.multi_playlist.append(
			self.tauon.pl_gen(title=self.strings.spotify_albums, playlist_ids=playlist)
		)
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "sal"
		self.spotify_com = False

	def append_track(self, url: str, playlist_number: int | None = None) -> None:
		self.connect()
		if not self.spotify:
			return

		if url.startswith("spotify:track:"):
			id = url[14:]
		else:
			url = url.split("?")[0]
			id = url.strip("/").split("/")[-1]

		if playlist_number is None:
			playlist_number = self.tauon.pctl.active_playlist_viewing

		track = self.spotify.track(id, market=self.country)
		tr = self.load_track(track)
		self.tauon.pctl.master_library[tr.index] = tr
		self.tauon.pctl.multi_playlist[playlist_number].playlist_ids.append(tr.index)
		self.tauon.gui.pl_update += 1

	def append_album(self, url: str, playlist_number: int | None = None, return_list: bool = False) -> list | None:
		self.connect()
		if not self.spotify:
			return []

		if url.startswith("spotify:album:"):
			id = url[14:]
		else:
			url = url.split("?")[0]
			id = url.strip("/").split("/")[-1]

		album = self.spotify.album(id)
		playlist: list[int] = []
		self.update_existing_import_list()
		self.load_album(album, playlist)

		if return_list:
			return playlist

		if playlist_number is None:
			playlist_number = self.tauon.pctl.active_playlist_viewing

		self.tauon.pctl.multi_playlist[playlist_number].playlist_ids.extend(playlist)
		self.tauon.gui.pl_update += 1
		return None

	def playlist(self, url: str, return_list: bool = False, silent: bool = False) -> list[int] | None:
		self.connect()
		if not self.spotify:
			return []

		if url.startswith("spotify:playlist:"):
			id = url[17:]
		else:
			url = url.split("?")[0]
			if len(url) != 22:
				id = url.strip("/").split("/")[-1]
			else:
				id = url

		if len(id) != 22:
			logging.error("ID Error")
			if return_list:
				return []
			return None

		p = self.spotify.playlist(id, market=self.country)
		playlist: list[int] = []
		self.update_existing_import_list()
		pages = self.spotify.all_pages(p.tracks)
		for page in pages:
			for item in page.items:
				if type(item.track) is tk.model.FullPlaylistTrack:
					nt = self.load_track(item.track, include_album_url=True)
					self.tauon.pctl.master_library[nt.index] = nt
					playlist.append(nt.index)

		if return_list:
			return playlist

		title = p.name + " by " + p.owner.display_name
		if p.name in ("Discover Weekly", "Release Radar"):
			# self.tauon.pctl.multi_playlist[len(self.tauon.pctl.multi_playlist) - 1].hide_title = True
			title = p.name
		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=title, playlist_ids=playlist))

		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = f'spl"{id}"'
		if not silent:
			self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)
		return None

	# def rec_playlist(self, artist_url: str, track_url: str) -> None:
	#
	# 	self.connect()
	# 	if not self.spotify:
	# 		return
	#
	# 	id = artist_url.strip("/").split("/")[-1]
	# 	track_ids = None
	# 	if track_url:
	# 		track_ids = [track_url.strip("/").split("/")[-1]]
	# 	artist = self.spotify.artist(id)
	# 	recs = self.spotify.recommendations(artist_ids=[id], track_ids=track_ids, limit=25, market=self.country)
	# 	playlist = []
	# 	self.update_existing_import_list()
	# 	for t in recs.tracks:
	# 		nt = self.load_track(t, update_master_count=True, include_album_url=True)
	# 		self.tauon.pctl.master_library[nt.index] = nt
	# 		playlist.append(nt.index)
	#
	# 	self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=artist.name + " Recs", playlist_ids=playlist))
	# 	self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)
	# 	self.tauon.gui.message_box = False

	def artist_playlist(self, url: str) -> None:
		self.connect()
		if not self.spotify:
			return
		id = url.strip("/").split("/")[-1]
		artist = self.spotify.artist(id)
		artist_albums = self.spotify.artist_albums(id, limit=50, include_groups=["album"])
		playlist: list[int] = []
		self.update_existing_import_list()

		for a in artist_albums.items:
			full_album = self.spotify.album(a.id)
			self.load_album(full_album, playlist)

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=artist.name, playlist_ids=playlist))
		self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)
		self.tauon.gui.message_box = False

		artist_albums = self.spotify.artist_albums(id, limit=50, include_groups=["single"])
		playlist: list[int] = []
		self.update_existing_import_list()

		for a in artist_albums.items:
			full_album = self.spotify.album(a.id)
			self.load_album(full_album, playlist)

		self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=artist.name + " Singles", playlist_ids=playlist))
		self.tauon.gui.message_box = False

	def album_playlist(self, url: str) -> None:
		l = self.append_album(url, return_list=True)
		self.tauon.pctl.multi_playlist.append(
			self.tauon.pl_gen(
				title=f"{self.tauon.pctl.get_track(l[0]).artist} - {self.tauon.pctl.get_track(l[0]).album}",
				playlist_ids=l,
				hide_title=False,
			),
		)
		self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)

	def update_existing_import_list(self) -> None:
		self.current_imports.clear()
		for tr in self.tauon.pctl.master_library.values():
			if "spotify-track-url" in tr.misc:
				self.current_imports[tr.misc["spotify-track-url"]] = tr

	def create_playlist(self, name: str) -> str | None:
		logging.info("Create new spotify playlist")
		self.connect()
		if not self.spotify:
			return None

		try:
			user = self.spotify.current_user()
			playlist = self.spotify.playlist_create(user.id, name, True)
			return playlist.id
		except Exception:
			logging.exception("Failed to create playlist")
			return None

	def upload_playlist(self, playlist_id: str, track_urls: list[str]) -> None:
		self.connect()
		if not self.spotify:
			return

		try:
			uris = []
			for url in track_urls:
				uris.append("spotify:track:" + url.strip("/").split("/")[-1])

			self.spotify.playlist_clear(playlist_id)
			time.sleep(0.05)
			with self.spotify.chunked(True):
				self.spotify.playlist_add(playlist_id, uris)
		except Exception:
			logging.exception("Spotify upload error!")
			self.show_message(_("Spotify upload error!"), mode="error")

	def load_album(self, album: FullAlbum, playlist: list[int] | None) -> None:
		# a = item
		album_url = album.external_urls["spotify"]
		art_url = album.images[0].url
		album_name = album.name
		total_tracks = album.total_tracks
		date = album.release_date
		album_artist = album.artists[0].name
		id = album.id
		parent = (album_artist + " - " + album_name).strip("- ")

		# logging.info(a.release_date, a.name)
		for track in album.tracks.items:
			pr = None
			if "spotify" in track.external_urls:
				pr = self.current_imports.get(track.external_urls["spotify"])
			if pr:
				new = False
				nt = pr
			else:
				new = True
				nt = self.tauon.TrackClass()
				nt.index = self.tauon.pctl.master_count

			nt.is_network = True
			nt.file_ext = "SPTY"
			nt.url_key = track.id
			if track.artists and "spotify" in track.artists[0].external_urls:
				nt.misc["spotify-artist-url"] = track.artists[0].external_urls["spotify"]
			nt.misc["spotify-album-url"] = album_url
			if "spotify" in track.external_urls:
				nt.misc["spotify-track-url"] = track.external_urls["spotify"]
			nt.artist = track.artists[0].name
			nt.album_artist = album_artist
			nt.date = date
			nt.album = album_name
			nt.disc_number = track.disc_number
			# nt.disc_total =
			nt.length = track.duration_ms / 1000
			nt.title = track.name
			nt.track_number = track.track_number
			nt.track_total = total_tracks
			nt.art_url_key = art_url
			nt.parent_folder_path = parent
			nt.parent_folder_name = parent
			if new:
				self.tauon.pctl.master_count += 1
				self.tauon.pctl.master_library[nt.index] = nt
			playlist.append(nt.index)

	def load_track(
		self, track: tk.model.FullPlaylistTrack, update_master_count: bool = True, include_album_url: bool = False
	) -> TrackClass:
		if "spotify" in track.external_urls:
			pr = self.current_imports.get(track.external_urls["spotify"])
		else:
			pr = False

		if pr:
			new = False
			nt = pr
		else:
			new = True
			nt = self.tauon.TrackClass()
			nt.index = self.tauon.pctl.master_count

		nt.found = False
		if track.is_playable is True or track.is_playable is None:
			nt.found = True
		nt.is_network = True
		nt.file_ext = "SPTY"
		nt.url_key = track.id
		# if new:
		if "spotify" in track.artists[0].external_urls:
			nt.misc["spotify-artist-url"] = track.artists[0].external_urls["spotify"]
		if include_album_url and "spotify-album-url" not in nt.misc:
			if "spotify" in track.album.external_urls:
				nt.misc["spotify-album-url"] = track.album.external_urls["spotify"]
		if "spotify" in track.external_urls:
			nt.misc["spotify-track-url"] = track.external_urls["spotify"]
		if track.artists[0].name:
			nt.artist = track.artists[0].name
		if track.album.artists:
			nt.album_artist = track.album.artists[0].name
		if track.album.release_date:
			nt.date = track.album.release_date
		nt.album = track.album.name
		nt.disc_number = track.disc_number
		nt.length = track.duration_ms / 1000
		nt.title = track.name
		nt.track_number = track.track_number
		# nt.track_total = total_tracks
		if track.album.images:
			nt.art_url_key = track.album.images[0].url
		parent = (nt.album_artist + " - " + nt.album).strip("- ")
		nt.parent_folder_path = parent
		nt.parent_folder_name = parent

		if update_master_count and new:
			self.tauon.pctl.master_count += 1

		return nt

	def like_track(self, tract_object: TrackClass) -> None:
		self.connect()
		if not self.spotify:
			return
		track_url = tract_object.misc.get("spotify-track-url", False)
		if track_url:
			id = track_url.strip("/").split("/")[-1]
			results = self.spotify.saved_tracks_contains([id])
			if not results or results[0] is False:
				self.spotify.saved_tracks_add([id])
				tract_object.misc["spotify-liked"] = True
				self.show_message(self.strings.spotify_like_added, mode="done")
				return
			self.show_message(self.strings.spotify_already_liked)
			return

	def unlike_track(self, tract_object: TrackClass) -> None:
		self.connect()
		if not self.spotify:
			return
		track_url = tract_object.misc.get("spotify-track-url", False)
		if track_url:
			id = track_url.strip("/").split("/")[-1]
			results = self.spotify.saved_tracks_contains([id])
			if not results or results[0] is True:
				self.spotify.saved_tracks_delete([id])
				tract_object.misc.pop("spotify-liked", None)
				self.show_message(self.strings.spotify_un_liked, mode="done")
				return
			self.show_message(self.strings.spotify_already_un_liked)
			return

	def get_library_likes(self, return_list: bool = False) -> list | None:
		self.connect()
		if not self.spotify:
			self.spotify_com = False
			self.show_message(self.strings.spotify_need_enable)
			return []

		self.update_existing_import_list()
		tracks = self.spotify.saved_tracks(market=self.country)

		playlist = []

		for tr in self.tauon.pctl.master_library.values():
			tr.misc.pop("spotify-liked", None)

		pages = self.spotify.all_pages(tracks)
		for page in pages:
			for item in page.items:
				nt = self.load_track(item.track)
				self.tauon.pctl.master_library[nt.index] = nt
				playlist.append(nt.index)
				nt.misc["spotify-liked"] = True

		if return_list:
			return playlist

		for p in self.tauon.pctl.multi_playlist:
			if p.title == self.tauon.strings.spotify_likes:
				p.playlist_ids[:] = playlist[:]
				self.spotify_com = False
				return None

		self.tauon.pctl.multi_playlist.append(
			self.tauon.pl_gen(title=self.tauon.strings.spotify_likes, playlist_ids=playlist)
		)
		self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "slt"
		self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)
		self.spotify_com = False
		return None

	def monitor(self) -> None:
		tr = self.tauon.pctl.playing_object()
		result = None

		# Detect if playback has resumed
		if self.playing and self.paused:
			result = self.spotify.playback_currently_playing()
			if result and result.is_playing:
				self.paused = False
				self.progress_timer.set()
				self.tauon.pctl.playing_state = 1
				self.tauon.gui.update += 1

		# Detect is playback has been modified
		elif self.playing and self.start_timer.get() > 4 and self.tauon.pctl.playing_time + 5 < tr.length:
			if not result:
				result = self.spotify.playback_currently_playing()

			# Playback has been stopped?
			if (result is None or result.item is None) or tr is None:
				self.playing = False
				self.tauon.pctl.stop()
				return
			# Playback has been paused?
			if tr and result and not result.is_playing:
				self.paused = True
				self.tauon.pctl.playing_state = 2
				self.tauon.gui.update += 1
				return
			# Something else is now playing? If so, switch into coast mode
			if result.item.name != tr.title:
				self.tauon.pctl.playing_state = 3
				self.playing = False
				self.coasting = True
				self.coast_update(result)
				self.tauon.gui.pl_update += 2
				return

			p = result.progress_ms
			if p is not None:
				# if abs(self.tauon.pctl.playing_time - (p / 1000)) > 0.4:
				# 	logging.info("DESYNC")
				# 	logging.info(abs(self.tauon.pctl.playing_time - (p / 1000)))
				self.tauon.pctl.playing_time = p / 1000
				self.tauon.pctl.decode_time = self.tauon.pctl.playing_time
				# else:
				# logging.info("SYNCED")

	def update(self, start: bool = False) -> None:
		if self.playing:
			self.coasting = False
			return

		self.connect()
		if not self.spotify:
			return

		result = self.spotify.playback_currently_playing()
		self.tauon.thread_manager.ready_playback()

		if self.playing or (not self.coasting and not start):
			return

		try:
			self.tauon.thread_manager.player_lock.release()
		except Exception:
			logging.exception("Failed to release lock")

		if result is None or result.is_playing is False:
			if self.coasting:
				if self.tauon.pctl.radio_image_bin:
					self.loaded_art = ""
					self.tauon.pctl.radio_image_bin.close()
					self.tauon.pctl.radio_image_bin = None
					self.tauon.dummy_track.artist = ""
					self.tauon.dummy_track.date = ""
					self.tauon.dummy_track.title = ""
					self.tauon.dummy_track.album = ""
					self.tauon.dummy_track.art_url_key = ""
					self.tauon.gui.clear_image_cache_next = True
					self.paused = True

			else:
				self.show_message(self.strings.spotify_not_playing)
			return

		self.coasting = True
		self.started_once = True
		self.tauon.pctl.playing_state = 3

		if result.is_playing:
			self.paused = False
		else:
			self.paused = True

		self.coast_update(result)

	def append_playing(self, playlist_number: int) -> None:
		if not self.coasting:
			return
		tr = self.tauon.pctl.playing_object()
		if tr and "spotify-track-url" in tr.misc:
			self.append_track(tr.misc["spotify-track-url"], playlist_number=playlist_number)

	def coast_update(self, result: dict | None) -> None:
		if result is None or result.item is None:
			logging.info("Spotify returned unknown")
			return

		self.tauon.dummy_track.artist = result.item.artists[0].name
		self.tauon.dummy_track.title = result.item.name
		self.tauon.dummy_track.album = result.item.album.name
		self.tauon.dummy_track.date = result.item.album.release_date
		self.tauon.dummy_track.file_ext = "Spotify"

		self.progress_timer.set()
		self.update_timer.set()

		d = result.item.duration_ms
		if d is not None:
			self.tauon.pctl.playing_length = d / 1000

		p = result.progress_ms
		if p is not None:
			self.tauon.pctl.playing_time = p / 1000

		self.tauon.pctl.decode_time = self.tauon.pctl.playing_time

		art_url = result.item.album.images[0].url
		self.tauon.dummy_track.url_key = result.item.id
		self.tauon.dummy_track.misc["spotify-album-url"] = result.item.album.external_urls["spotify"]
		self.tauon.dummy_track.misc["spotify-track-url"] = result.item.external_urls["spotify"]

		if art_url and self.loaded_art != art_url:
			self.loaded_art = art_url
			art_response = requests.get(art_url, timeout=10)
			if self.tauon.pctl.radio_image_bin:
				self.tauon.pctl.radio_image_bin.close()
				self.tauon.pctl.radio_image_bin = None
			self.tauon.pctl.radio_image_bin = io.BytesIO(art_response.content)
			self.tauon.pctl.radio_image_bin.seek(0)
			self.tauon.dummy_track.art_url_key = "ok"
			self.tauon.gui.clear_image_cache_next = True

		self.tauon.gui.update += 2
		self.tauon.gui.pl_update += 1

	def import_context(self) -> None:
		self.connect()
		if not self.spotify:
			return
		result = self.spotify.playback_currently_playing()
		if not result or not result.context:
			self.show_message(_("No Spotify context found"))
			return

		if result.context.type == "playlist":
			self.playlist(result.context.uri)

		if result.context.type == "album":
			self.album_playlist(result.context.uri)

		if result.context.type == "artist":
			self.artist_playlist(result.context.uri)

		self.tauon.gui.pl_update += 1



"""Tauon Music Box - Subsonic/Airsonic Integration"""

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import hashlib
import io
import json
import logging
import math
import os
import re
import secrets
import threading
import time
from typing import TYPE_CHECKING

import requests

from tauon.t_modules.t_extra import StarRecord

if TYPE_CHECKING:
	from io import BytesIO
	from typing import Any

	from tauon.t_modules.t_main import AlbumStarStore, GuiVar, PlayerCtl, StarStore, Tauon, TrackClass
	from tauon.t_modules.t_prefs import Prefs


class SubsonicService:
	def __init__(self, tauon: Tauon, album_star_store: AlbumStarStore) -> None:
		self.tauon: Tauon = tauon
		self.gui: GuiVar = tauon.gui
		self.pctl: PlayerCtl = tauon.pctl
		self.prefs: Prefs = tauon.prefs
		self.t_title: str = tauon.t_title
		self.star_store: StarStore = tauon.star_store
		self.album_star_store: AlbumStarStore = album_star_store
		self.show_message = tauon.show_message
		self.playlists = tauon.prefs.subsonic_playlists
		self.scanning: bool = False


	def _as_list(self, value):
		"""Subsonic can return a dict instead of a list for single-item collections."""
		if value is None:
			return []
		if isinstance(value, list):
			return value
		return [value]

	def r(self, point: str, p: dict[str, str] | None = None, binary: bool = False, get_url: bool = False):
		salt = secrets.token_hex(8)
		server = self.prefs.subsonic_server.rstrip("/") + "/"

		params = {
			"u": self.prefs.subsonic_user,
			"v": "1.13.0",
			"c": self.t_title,
			"f": "json",
		}

		if self.prefs.subsonic_password_plain:
			params["p"] = self.prefs.subsonic_password
		else:
			params["t"] = hashlib.md5((self.prefs.subsonic_password + salt).encode()).hexdigest()
			params["s"] = salt

		if p:
			params.update(p)

		point = "rest/" + point

		url = server + point

		if get_url:
			return url, params

		response = requests.get(url, params=params, timeout=10)

		if binary:
			return response.content

		# Some broken servers can send invalid JSON with control chars - remove them, see https://github.com/Taiko2k/Tauon/issues/1112
		CONTROL_CHAR_RE = re.compile(r"[\x00-\x08\x0B\x0C\x0E-\x1F]")
		control_chars = CONTROL_CHAR_RE.findall(response.text)
		if control_chars:
			clean_response = CONTROL_CHAR_RE.sub("", response.text)
			details = [f"U+{ord(c):04X}" for c in control_chars]
			logging.warning(f"Invalid control characters found in JSON response: {', '.join(details)}")
		else:
			clean_response = response.text

		try:
			d = json.loads(clean_response)
		except json.decoder.JSONDecodeError:
			logging.exception(f"Failed to decode subsonic response as json: {clean_response}")
			return None
		except Exception:
			logging.exception(f"Unknown error loading subsonic response: {clean_response}")
			return None
		# logging.info(d)

		if d["subsonic-response"]["status"] != "ok":
			self.show_message(_("Subsonic Error: ") + response.text, mode="warning")
			logging.error(f"Subsonic Error: {response.text}")

		return d

	def get_cover(self, track_object: TrackClass) -> BytesIO:
		response = self.r("getCoverArt", p={"id": track_object.art_url_key}, binary=True)
		try:
			response.decode("utf-8")
			raise ValueError(f"Expected binary data with an image but got a valid string: {response}")
		except UnicodeDecodeError:
			pass

		return io.BytesIO(response)

	def resolve_stream(self, key: str) -> tuple[str, dict[str, str]] | bytes | Any | None:
		p = {"id": key}
		if self.prefs.network_stream_bitrate > 0:
			p["maxBitRate"] = self.prefs.network_stream_bitrate

		return self.r("stream", p=p, get_url=True)
		# logging.info(response.content)

	def listen(self, track_object: TrackClass, submit: bool = False) -> bool:

		try:
			a = self.r("scrobble", p={"id": track_object.url_key, "submission": submit})
		except Exception:
			logging.exception("Error connecting for scrobble on airsonic")
		return True

	def set_rating(self, track_object: TrackClass, rating: int) -> bool:
		try:
			a = self.r("setRating", p={"id": track_object.url_key, "rating": math.ceil(rating / 2)})
		except Exception:
			logging.exception("Error connect for set rating on airsonic")
		return True

	def set_album_rating(self, track_object: TrackClass, rating: int) -> bool:
		id = track_object.misc.get("subsonic-folder-id")
		if id is not None:
			try:
				a = self.r("setRating", p={"id": id, "rating": math.ceil(rating / 2)})
			except Exception:
				logging.exception("Error connect for set rating on airsonic")
		return True

	def star_track(self, track_object: TrackClass) -> bool:
		try:
			response = self.r("star", p={"id": track_object.url_key})
		except Exception:
			logging.exception("Error connect for star track on airsonic")
			return False

		if not response:
			logging.error(f"Airsonic star request returned no response for track id={track_object.url_key!r}")
			return False

		subsonic_response = response.get("subsonic-response", {})
		if subsonic_response.get("status") != "ok":
			error = subsonic_response.get("error", {})
			logging.error(
				"Airsonic star request failed for track id=%r code=%r message=%r",
				track_object.url_key,
				error.get("code"),
				error.get("message"),
			)
			return False

		return True

	def unstar_track(self, track_object: TrackClass) -> bool:
		try:
			response = self.r("unstar", p={"id": track_object.url_key})
		except Exception:
			logging.exception("Error connect for unstar track on airsonic")
			return False

		if not response:
			logging.error(f"Airsonic unstar request returned no response for track id={track_object.url_key!r}")
			return False

		subsonic_response = response.get("subsonic-response", {})
		if subsonic_response.get("status") != "ok":
			error = subsonic_response.get("error", {})
			logging.error(
				"Airsonic unstar request failed for track id=%r code=%r message=%r",
				track_object.url_key,
				error.get("code"),
				error.get("message"),
			)
			return False

		return True

	def star_album(self, track_object: TrackClass) -> bool:
		id = track_object.misc.get("subsonic-folder-id")
		if id is not None:
			try:
				a = self.r("star", p={"id": id})
			except Exception:
				logging.exception("Error connect for star Album on airsonic")
		return True

	def get_music3(self, return_list: bool = False) -> list[int] | None:
		self.scanning = True
		self.gui.to_got = 0

		existing: dict[str, int] = {}

		for track_id, track in self.pctl.master_library.items():
			if track.is_network and track.file_ext == "SUB":
				existing[track.url_key] = track_id

		try:
			a = self.r("getIndexes")
		except Exception:
			logging.exception("Error connecting to Airsonic server")
			self.show_message(_("Error connecting to Airsonic server"), mode="error")
			self.scanning = False
			return []
		if not a or "subsonic-response" not in a:
			logging.error("Invalid response from Airsonic getIndexes")
			self.show_message(_("Error connecting to Airsonic server"), mode="error")
			self.scanning = False
			return []

		# {'openSubsonic': True, 'serverVersion': '8', 'status': 'failed', 'type': 'lms', 'version': '1.16.0', 'error': {'code': 41, 'message': 'Token authentication not supported for LDAP users.'}}
		if "indexes" not in a["subsonic-response"]:
			self.scanning = False
			if "error" in a["subsonic-response"]:
				logging.debug(a["subsonic-response"])
				self.show_message(
					_("Error connecting to Airsonic server"),
					f"{a['subsonic-response']['error']['code']}: {a['subsonic-response']['error']['message']}",
					mode="error",
				)
				return None
			logging.critical("Failed to find expected key 'indexes', report a bug with the log below!")
			logging.critical(a["subsonic-response"])
			self.show_message(
				_("Error connecting to Airsonic server"), "See console log for more details", mode="error"
			)
			return None

		indexes = self._as_list(a["subsonic-response"]["indexes"].get("index"))

		folders: list[tuple[str, str]] = []

		for letter in indexes:
			artists = self._as_list(letter.get("artist"))
			for artist in artists:
				folders.append(
					(
						artist["id"],
						artist["name"],
					)
				)

		playlist: list[int] = []
		songsets: list[tuple[TrackClass, str, str, int]] = []
		for i in range(len(folders)):
			songsets.append([])
		statuses = [0] * len(folders)
		liked_track_ids: list[int] = []

		def getsongs(
			index: int, folder_id: str, name: str, inner: bool = False, parent: dict[str, str | int] | None = None
		) -> None:
			try:
				d = self.r("getMusicDirectory", p={"id": folder_id})
				if "child" not in d["subsonic-response"]["directory"]:
					if not inner:
						statuses[index] = 2
					return

			except json.decoder.JSONDecodeError:
				logging.exception("Error reading Airsonic directory")
				if not inner:
					statuses[index] = 2
				self.show_message(_("Error reading Airsonic directory!"), mode="warning")
				return
			except Exception:
				logging.exception("Unknown Error reading Airsonic directory")
				if not inner:
					statuses[index] = 2
				return

			items = self._as_list(d["subsonic-response"]["directory"].get("child"))

			self.gui.update = 2

			for item in items:
				if item.get("isDir"):
					if "userRating" in item and "artist" in item:
						rating = item["userRating"]
						if (
							self.album_star_store.get_rating_artist_title(item["artist"], item["title"]) == 0
							and rating == 0
						):
							pass
						else:
							self.album_star_store.set_rating_artist_title(
								item["artist"], item["title"], int(rating * 2)
							)

					getsongs(index, item["id"], item["title"], inner=True, parent=item)
					continue

				self.gui.to_got += 1
				song = item
				nt = self.tauon.TrackClass()

				if parent and "artist" in parent:
					nt.album_artist = parent["artist"]

				if "title" in song:
					nt.title = song["title"]
				if "artist" in song:
					nt.artist = song["artist"]
				if "album" in song:
					nt.album = song["album"]
				if "track" in song:
					nt.track_number = song["track"]
				if "year" in song:
					nt.date = str(song["year"])
				if "duration" in song:
					nt.length = song["duration"]

				nt.file_ext = "SUB"
				nt.parent_folder_name = name
				if "path" in song:
					nt.fullpath = song["path"]
					nt.parent_folder_path = os.path.dirname(song["path"])
				if "coverArt" in song:
					nt.art_url_key = song["coverArt"]
				nt.url_key = song["id"]
				nt.misc["subsonic-folder-id"] = folder_id
				nt.is_network = True

				rating = 0
				if "userRating" in song:
					rating = int(song["userRating"])

				songsets[index].append((nt, name, song["id"], rating))

			if inner:
				return
			statuses[index] = 2

		i = -1
		for folder_id, name in folders:
			i += 1
			while statuses.count(1) > 3:
				time.sleep(0.1)

			statuses[i] = 1
			t = threading.Thread(target=getsongs, args=([i, folder_id, name]))
			t.daemon = True
			t.start()

		while statuses.count(2) != len(statuses):
			time.sleep(0.1)

		try:
			a = self.r("getStarred2")
		except Exception:
			logging.exception("Error connecting to Airsonic server")
			self.show_message(_("Error connecting to Airsonic server"), mode="error")
			self.scanning = False
			return []

		liked_tracks: list[str] = []
		starred2 = a.get("subsonic-response", {}).get("starred2", {})
		for song in self._as_list(starred2.get("song")):
			song_id = song.get("id")
			if song_id:
				liked_tracks.append(song_id)

		for sset in songsets:
			for nt, name, song_id, rating in sset:
				id = self.pctl.master_count

				replace_existing = False
				ex = existing.get(song_id)
				if ex is not None:
					id = ex
					replace_existing = True

				nt.index = id
				self.pctl.master_library[id] = nt
				if not replace_existing:
					self.pctl.master_count += 1

				if nt.url_key in liked_tracks:
					liked_track_ids.append(nt.index)

				playlist.append(nt.index)

				if self.star_store.get_rating(nt.index) == 0 and rating == 0:
					pass
				else:
					self.star_store.set_rating(nt.index, rating * 2)

		def set_favs(d: list[int]) -> None:
			for track_id in d:
				if track_id == 0:
					continue

				star = self.tauon.star_store.full_get(track_id)
				if star is None:
					star = StarRecord()
				if not star.loved:
					star.loved = True
				self.tauon.star_store.insert(track_id, star)

		set_favs(liked_track_ids)

		self.scanning = False
		if return_list:
			return playlist

		self.pctl.multi_playlist.append(self.tauon.pl_gen(title=_("Airsonic Collection"), playlist_ids=playlist))
		self.pctl.gen_codes[self.pctl.pl_to_id(len(self.pctl.multi_playlist) - 1)] = "air"
		self.pctl.switch_playlist(len(self.pctl.multi_playlist) - 1)
		return None
